# Copyright The IETF Trust 2017, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

series_name = {
    'std':  'STD',
    'bcp':  'BCP',
    'info': 'FYI',
}

category_name = {
    'rfc':      'Request for Comments',
    'std':      'Standards Track',
    'bcp':      'Best Current Practice',
    'exp':      'Experimental',
    'info':     'Informational',
    'historic': 'Historic',
}

stream_name = {
    'IETF':         'Internet Engineering Task Force (IETF)',
    'IAB':          'Internet Architecture Board (IAB)',
    'IRTF':         'Internet Research Task Force (IRTF)',
    'independent':  'Independent Submission',
    'editorial':    'Editorial Stream',
}
