# Copyright The IETF Trust 2017, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

# The following text entries were taken from
# https://www.ietf.org/ietf-ftp/1id-guidelines.txt

boilerplate_draft_status_of_memo = [
        """<t>
        This Internet-Draft is submitted in full conformance with the
        provisions of BCP 78 and BCP 79.
        </t>""",
        """<t>
        Internet-Drafts are working documents of the Internet Engineering Task
        Force (IETF). Note that other groups may also distribute working
        documents as Internet-Drafts. The list of current Internet-Drafts is
        at <eref target="{scheme}://datatracker.ietf.org/drafts/current/"/>.
        </t>""",
        """<t>
        Internet-Drafts are draft documents valid for a maximum of six months
        and may be updated, replaced, or obsoleted by other documents at any
        time. It is inappropriate to use Internet-Drafts as reference
        material or to cite them other than as "work in progress."
        </t>""",
        """<t>
        This Internet-Draft will expire on {expiration_date}.
        </t>""",
    ]
